"""Constants for the Rollease Acmeda Automate integration."""
import logging

LOGGER = logging.getLogger(__package__)
DOMAIN = "acmeda"

ACMEDA_HUB_UPDATE = "acmeda_hub_update_{}"
ACMEDA_ENTITY_REMOVE = "acmeda_entity_remove_{}"
