"""Constants for the IPP integration."""

# Integration domain
DOMAIN = "ipp"

# Attributes
ATTR_COMMAND_SET = "command_set"
ATTR_IDENTIFIERS = "identifiers"
ATTR_INFO = "info"
ATTR_LOCATION = "location"
ATTR_MANUFACTURER = "manufacturer"
ATTR_MARKER_TYPE = "marker_type"
ATTR_MARKER_LOW_LEVEL = "marker_low_level"
ATTR_MARKER_HIGH_LEVEL = "marker_high_level"
ATTR_MODEL = "model"
ATTR_SERIAL = "serial"
ATTR_SOFTWARE_VERSION = "sw_version"
ATTR_STATE_MESSAGE = "state_message"
ATTR_STATE_REASON = "state_reason"
ATTR_URI_SUPPORTED = "uri_supported"

# Config Keys
CONF_BASE_PATH = "base_path"
CONF_SERIAL = "serial"
CONF_TLS = "tls"
CONF_UUID = "uuid"
