"""Constants for launch_library."""

ATTR_AGENCY = "agency"
ATTR_AGENCY_COUNTRY_CODE = "agency_country_code"
ATTR_LAUNCH_TIME = "launch_time"
ATTR_STREAM = "stream"

ATTRIBUTION = "Data provided by Launch Library."

DEFAULT_NAME = "Next launch"
